import os

import yaml


class Configuration:
    def __init__(self):
        filepath = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'config.yaml')
        with open(filepath, 'r') as config_file:
            self._config = yaml.safe_load(config_file)

    def get(self, key: str):
        return self._config[key]


config = Configuration()



import logging.handlers
import os

INITIALIZED = False
if not INITIALIZED:
    logger = logging.getLogger('dragoneye')
    logger.propagate = False
    logger.setLevel(logging.INFO)
    logger.addHandler(logging.StreamHandler())
    INITIALIZED = True


def add_file_handler(logger_directory) -> bool:
    try:
        os.makedirs(logger_directory, exist_ok=True)
        logger_file = os.path.join(logger_directory, 'dragoneye.log')
        rotate_file = os.path.isfile(logger_file)
        file_handler = logging.handlers.RotatingFileHandler(logger_file, maxBytes=10000000, backupCount=10)
        file_handler.setFormatter(logging.Formatter('%(asctime)s %(name)-12s %(levelname)-8s %(message)s'))
        if rotate_file:  # keeping log file per execution
            file_handler.doRollover()
        logger.addHandler(file_handler)
        return True
    except Exception:
        return False



import json
import subprocess
import sys
from typing import Optional

import requests
from dragoneye.utils.app_logger import logger
from dragoneye.dragoneye_exception import DragoneyeException
from dragoneye.utils.misc_utils import invoke_get_request
from dragoneye.utils.value_validator import validate_uuid


class AzureAuthorizer:

    @staticmethod
    def get_authorization_token(subscription_id: str,
                                tenant_id: Optional[str] = None,
                                client_id: Optional[str] = None,
                                client_secret: Optional[str] = None) -> str:
        """
        This function generates a JWT bearer token.

        If client_id and client_secret are provided, it will generate the token using these credentials.

        Otherwise, it will attempt to generate a token from your CLI credentials, using
        `az account get-access-token <https://docs.microsoft.com/en-us/cli/azure/account?view=azure-cli-latest#az_account_get_access_token>`__
        """
        validate_uuid(subscription_id, 'Invalid subscription id')
        if not (client_id and client_secret and tenant_id):
            token = AzureAuthorizer._get_token_from_az_cli()
        else:
            token = AzureAuthorizer._get_token_from_credentials(tenant_id, client_id, client_secret)

        auth_header = f'Bearer {token}'
        AzureAuthorizer._test_connectivity(subscription_id, auth_header)

        logger.info('JWT bearer token generated successfully')
        return auth_header

    @staticmethod
    def _get_token_from_credentials(tenant_id: str, client_id: str, client_secret: str) -> str:
        validate_uuid(tenant_id, 'Invalid tenant id')
        validate_uuid(client_id, 'Invalid client id')

        logger.info('Will try to generate JWT bearer token using provided client id/secret...')
        response = requests.post(
            url=f'https://login.microsoftonline.com/{tenant_id}/oauth2/token',
            data={
                'grant_type': 'client_credentials',
                'client_id': client_id,
                'client_secret': client_secret,
                'resource': 'https://management.azure.com/'
            }
        )

        if response.status_code != 200:
            raise DragoneyeException(f'Failed to authenticate. status code: {response.status_code}\n'
                                     f'Reason: {response.text}', response.text)

        response_body = json.loads(response.text)
        access_token = response_body['access_token']
        return access_token

    @staticmethod
    def _get_token_from_az_cli() -> str:
        logger.info('Will try to generate JWT bearer token from currently logged in azure user...')
        with subprocess.Popen(['az', 'account', 'get-access-token'],
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE) as process:
            stdout, stderr = process.communicate()
            process.__enter__()
            if stderr:
                error = stderr.decode(sys.stderr.encoding)
                raise DragoneyeException('Failed to authenticate.\n'
                                         f'Reason: {error}', error)
            output = stdout.decode(sys.stdout.encoding)
            ind = output.rindex('}') + 1
            output = output[:ind]
            return json.loads(output)['accessToken']

    @staticmethod
    def _test_connectivity(subscription_id, token):
        headers = {
            'Authorization': token
        }
        url = f'https://management.azure.com/subscriptions/{subscription_id}/resourcegroups?api-version=2020-09-01'
        response = invoke_get_request(url, headers)
        if response.status_code != 200:
            raise DragoneyeException(f'Failed to authenticate. status code: {response.status_code}\n'
                                     f'Reason: {response.text}', response.text)



from typing import Sequence

from google.auth.exceptions import RefreshError
from google.oauth2 import service_account
from google.auth import impersonated_credentials, aws
from googleapiclient.discovery import build
from oauth2client.client import GoogleCredentials

from dragoneye.dragoneye_exception import DragoneyeException
from dragoneye.utils.app_logger import logger


class GcpCredentialsFactory:
    @classmethod
    def from_service_account_info(cls, service_account_info: dict):
        logger.info('Will try to generate credentials from service account info...')
        credentials = service_account.Credentials.from_service_account_info(
            service_account_info)

        cls.test_connectivity(credentials)
        logger.info('Generated credentials successfully')
        return credentials

    @classmethod
    def get_default_credentials(cls):
        logger.info('Will try to generate the default credentials...')
        credentials = GoogleCredentials.get_application_default()

        cls.test_connectivity(credentials)
        logger.info('Generated credentials successfully')
        return credentials

    @classmethod
    def from_service_account_file(cls, service_account_file: str):
        logger.info('Will try to generate credentials from service account file...')
        credentials = service_account.Credentials.from_service_account_file(
            service_account_file)

        cls.test_connectivity(credentials)
        logger.info('Generated credentials successfully')
        return credentials

    @classmethod
    def impersonate(cls, credentials, email: str, scopes: Sequence[str]):
        """
        Creates and returns impersonated credentials.
        :param credentials: The credentials of the source service account
        :param email: The email address of the target service account
        :param scopes: The scopes of the created credentials
        :return: An impersonated credentials instance
        """
        logger.info(f'Will try to generate short-lived credentials for {email}...')
        credentials = impersonated_credentials.Credentials(source_credentials=credentials,
                                                           target_principal=email,
                                                           target_scopes=scopes)
        cls.test_connectivity(credentials)
        logger.info('Generated credentials successfully')
        return credentials

    @classmethod
    def from_aws_credentials_config_info(cls, info: dict):
        logger.info('Will try to generate short-lived credentials for an AWS workload identity user from credentials config info...')
        cred = aws.Credentials.from_info(info)
        cls.test_connectivity(cred)
        logger.info('Generated credentials successfully')
        return cred

    @classmethod
    def from_aws_credentials_config_file(cls, config_file_path: str):
        logger.info('Will try to generate short-lived credentials for an AWS workload identity user from credentials config file...')
        cred = aws.Credentials.from_file(config_file_path)
        cls.test_connectivity(cred)
        logger.info('Generated credentials successfully')
        return cred

    @staticmethod
    def test_connectivity(credentials):
        with build('compute', 'v1', credentials=credentials) as service:
            try:
                service.instances().get(project='abc', zone='us-east1-a', instance='abc').execute()
            except RefreshError as ex:
                raise DragoneyeException('Unable to invoke GCP API with given credentials', str(ex))
            except Exception:
                pass



import os

from dragoneye.cloud_scanner.base_cloud_scanner import CloudScanSettings, CloudProvider


class GcpCloudScanSettings(CloudScanSettings):
    def __init__(self,
                 commands_path: str,
                 account_name: str,
                 project_id: str,
                 output_path: str = os.getcwd(),
                 should_clean_before_scan: bool = True):
        """
        The settings that the AwsScanner uses for aws scanning.

            :param commands_path: The path of a YAML file that describes the scan commands to be used.
            :param account_name: A name for the scan results.
            :param project_id: The project ID of the project to scan.
            :param output_path: The directory where results will be saved. Defaults to current working directory.
            :param should_clean_before_scan: A flag that determines if prior results of this specific account (identified by account_name)
                should be deleted before scanning.
        """
        super().__init__(CloudProvider.GCP, account_name, should_clean_before_scan, output_path, commands_path)
        self.project_id: str = project_id



import os
from typing import Optional

import click
from click_aliases import ClickAliasedGroup

from dragoneye.cloud_scanner.gcp.gcp_credentials_factory import GcpCredentialsFactory
from dragoneye.cloud_scanner.gcp.gcp_scan_settings import GcpCloudScanSettings
from dragoneye.cloud_scanner.gcp.gcp_scanner import GcpScanner
from dragoneye.version import __version__
from dragoneye.cloud_scanner.aws.aws_scanner import AwsScanner
from dragoneye.cloud_scanner.aws.aws_session_factory import AwsSessionFactory
from dragoneye.cloud_scanner.azure.azure_scanner import AzureScanner
from dragoneye.cloud_scanner.azure.azure_authorizer import AzureAuthorizer
from dragoneye.cloud_scanner.aws.aws_scan_settings import AwsCloudScanSettings
from dragoneye.cloud_scanner.azure.azure_scan_settings import AzureCloudScanSettings
from dragoneye.utils.value_validator import validate_uuid, validate_path


@click.group(name='scan',
             help='Execute data-fetching API commands on a cloud account at a large scale quickly. Currently supported: AWS, Azure, GCP',
             cls=ClickAliasedGroup)
@click.version_option(__version__)
def scan_cli():
    pass


def safe_cli_entry_point():
    try:
        scan_cli()
    except Exception as ex:
        click.echo(ex)


@scan_cli.command(name='gcp',
                  short_help='Scan a Google cloud provider account',
                  help='Scan a Google cloud provider account. '
                       '\n\nSCAN_COMMANDS_PATH: The file path to the yaml file that contains all the scan commands to run')
@click.argument('scan-commands-path',
                type=click.STRING)
@click.argument('project-id',
                type=click.STRING)
@click.option('--cloud-account-name', '-n',
              help='The name of your cloud account, the default value is \'default\'',
              type=click.STRING,
              default='default')
@click.option('--project-id',
              help='The ID of the project to scan',
              type=click.STRING,)
@click.option('--clean',
              help='Remove any existing data for the account before gathering',
              is_flag=True,
              default=True)
@click.option('--output-path',
              help='The path in which the scan results will be saved on. Defaults to current working directory.',
              type=click.STRING,
              default=os.getcwd())
@click.option('--credentials-path',
              help='The path to the `Google Application Credentials` json file. If left empty, will attempt to get the default credentials',
              type=click.STRING,
              default=None)
def gcp(scan_commands_path: str, project_id: str, clean: bool, output_path: str, cloud_account_name: str, credentials_path: Optional[str]):
    validate_path(scan_commands_path, f'Could not find file: {scan_commands_path}')

    gcp_scan_settings = GcpCloudScanSettings(commands_path=scan_commands_path,
                                             account_name=cloud_account_name,
                                             output_path=output_path,
                                             should_clean_before_scan=clean,
                                             project_id=project_id)
    if credentials_path:
        validate_path(credentials_path, f'Could not find file: {credentials_path}')
        credentials = GcpCredentialsFactory.from_service_account_file(credentials_path)
    else:
        credentials = GcpCredentialsFactory.get_default_credentials()

    output_path = GcpScanner(credentials, gcp_scan_settings).scan()
    click.echo(f'Results saved to {output_path}')


@scan_cli.command(name='azure',
                  short_help='Scan an Azure cloud account',
                  help='Scan an Azure cloud account. '
                       '\n\nSCAN_COMMANDS_PATH: The file path to the yaml file that contains all the scan commands to run')
@click.argument('scan-commands-path',
                type=click.STRING)
@click.option('--cloud-account-name', '-n',
              help='The name of your cloud account, the default value is \'default\'',
              type=click.STRING,
              default='default')
@click.option('--subscription-id', '-i',
              help='ID of Azure subscription to be added',
              type=click.STRING,
              required=True)
@click.option('--tenant-id', '-t',
              help='ID of Azure tenant of this subscription. Specify it only if you authenticate with client id/secret',
              type=click.STRING)
@click.option('--client-id', '-c',
              help='The client id of an application created in Azure. If not specified, will use credentials from `az login`',
              type=click.STRING)
@click.option('--client-secret', '-s',
              help='The client secret of an application created in Azure. If not specified, will use credentials from `az login`',
              type=click.STRING)
@click.option('--clean',
              help='Remove any existing data for the account before gathering',
              is_flag=True,
              default=True)
@click.option('--output-path',
              help='The path in which the scan results will be saved on. Defaults to current working directory.',
              type=click.STRING,
              default=os.getcwd())
def azure(cloud_account_name: str,
          subscription_id: str, client_id: str, client_secret: str, tenant_id: str,
          scan_commands_path, clean, output_path):
    validate_uuid(subscription_id, 'Invalid subscription id')
    validate_path(scan_commands_path, f'Could not find file: {scan_commands_path}')

    azure_scan_settings = AzureCloudScanSettings(
        commands_path=scan_commands_path,
        account_name=cloud_account_name,
        subscription_id=subscription_id,
        should_clean_before_scan=clean,
        output_path=output_path)

    auth_header = AzureAuthorizer.get_authorization_token(subscription_id, tenant_id, client_id, client_secret)
    output_path = AzureScanner(auth_header, azure_scan_settings).scan()
    click.echo(f'Results saved to {output_path}')


@scan_cli.command(name='aws',
                  short_help='Scan an AWS cloud account',
                  help='Scan an AWS cloud account. \n\nSCAN_COMMANDS_PATH: The file path to the yaml file that contains all the scan commands to run')
@click.argument('scan-commands-path',
                type=click.STRING)
@click.option('--cloud-account-name', '-n',
              help='The name of your cloud account, the default value is \'default\'',
              type=click.STRING,
              default='default')
@click.option('--profile',
              help='aws profile',
              type=click.STRING)
@click.option('--regions',
              help='Filter and query AWS only for the given regions (comma separated)',
              type=click.STRING,
              default='')
@click.option('--clean',
              help='Remove any existing data for the account before gathering',
              is_flag=True,
              default=True)
@click.option('--output-path',
              help='The path in which the scan results will be saved on. Defaults to current working directory.',
              type=click.STRING,
              default=os.getcwd())
@click.option('--default-region',
              help='The default region for scanning universal services. Defaults to the value of the AWS_DEFAULT_REGION environment variable.',
              type=click.STRING)
def aws(cloud_account_name,
        profile,
        regions,
        scan_commands_path,
        clean,
        output_path,
        default_region):
    aws_scan_settings = AwsCloudScanSettings(
        commands_path=scan_commands_path,
        account_name=cloud_account_name,
        regions_filter=regions.split(','),
        should_clean_before_scan=clean,
        output_path=output_path,
        default_region=default_region)

    validate_path(scan_commands_path, f'Could not find file: {scan_commands_path}')
    session = AwsSessionFactory.get_session(profile, default_region)
    output_path = AwsScanner(session, aws_scan_settings).scan()
    click.echo(f'Results saved to {output_path}')


if __name__ == '__main__':
    safe_cli_entry_point()



import functools
import time

from botocore.exceptions import ClientError

from dragoneye.utils.app_logger import logger

RATE_ERRORS = ['Throttling', 'TooManyRequestsException']


def rate_limiter(max_attempts=10):
    def decorator(func):
        @functools.wraps(func)
        def decorated(*args, **kwargs):
            attempt = 0
            while True:
                try:
                    return func(*args, **kwargs)
                except ClientError as ex:
                    if ex.response['Error']['Code'] not in RATE_ERRORS and ex.response['Error']['Message'] != 'Rate exceeded':
                        raise ex  # Raise non-throttling errors
                    attempt += 1
                    logger.warning(f'Throttling error on operation: {ex.operation_name} on attempt #{attempt}/{max_attempts}')
                    if attempt >= max_attempts:
                        raise ex
                    sleep_period = (2 ** attempt) - 1
                    time.sleep(sleep_period)

        return decorated
    return decorator



import collections
import os
from typing import List, Deque

import json

from requests import Response

from dragoneye.cloud_scanner.azure.azure_scan_settings import AzureCloudScanSettings
from dragoneye.cloud_scanner.base_cloud_scanner import BaseCloudScanner
from dragoneye.config import config
from dragoneye.utils.misc_utils import elapsed_time, invoke_get_request, init_directory, get_dynamic_values_from_files, custom_serializer
from dragoneye.utils.app_logger import logger
from dragoneye.utils.threading_utils import ThreadedFunctionData, execute_parallel_functions_in_threads


class AzureScanner(BaseCloudScanner):

    def __init__(self, auth_header: str, settings: AzureCloudScanSettings):
        super().__init__(settings)
        self.auth_header = auth_header
        self.subscription_id = settings.subscription_id

    @elapsed_time('Scanning Azure live environment took {} seconds')
    def scan(self) -> str:
        headers = {
            'Authorization': self.auth_header
        }

        self.account_data_dir = init_directory(self.settings.output_path, self.settings.account_name, self.settings.clean)
        resource_groups = self._get_resource_groups(headers)

        dependent_commands, independent_commands = self._get_scan_commands()

        non_dependable_tasks: List[ThreadedFunctionData] = []
        dependable_tasks: List[ThreadedFunctionData] = []
        deque_tasks: Deque[List[ThreadedFunctionData]] = collections.deque()

        for independent_command in independent_commands:
            non_dependable_tasks.append(ThreadedFunctionData(
                self._execute_scan_commands,
                (independent_command, headers, resource_groups),
                'exception on command {}'.format(independent_command)))

        deque_tasks.append(non_dependable_tasks)

        for dependent_command in dependent_commands:
            dependable_tasks.append(ThreadedFunctionData(
                self._execute_scan_commands,
                (dependent_command, headers, resource_groups),
                'exception on command {}'.format(dependent_command)))

        for dependable_task in dependable_tasks:
            deque_tasks.append([dependable_task])
        execute_parallel_functions_in_threads(deque_tasks, config.get('MAX_WORKERS'))

        self._print_summary()

        return os.path.abspath(os.path.join(self.account_data_dir, '..'))

    def _execute_scan_commands(self, scan_command: dict, headers: dict, resource_groups: List[str]) -> None:
        try:
            output_file = self._get_result_file_path(self.account_data_dir, scan_command['Name'])
            if os.path.isfile(output_file):
                # Data already scanned, so skip
                logger.warning('Response already present at {}'.format(output_file))
                return

            request = scan_command['Request']
            parameters = scan_command.get('Parameters', [])
            base_url = request.replace('{subscriptionId}', self.subscription_id)
            results = self._get_results(base_url, headers, parameters, self.account_data_dir, resource_groups)
            self._save_result(results, output_file)
            for url in results['urls']:
                logger.info(f'Results from {url} were saved to {output_file}')
        except Exception as ex:
            logger.exception('Exception occurred: {} while running command {}'.format(ex, scan_command))

    def _save_result(self, result: dict, filepath: str) -> None:
        self._add_resource_group(result)
        with open(filepath, "w") as file:
            json.dump(result, file, indent=4, default=custom_serializer)

    @staticmethod
    def _build_urls(_url: str, parameters: List[dict], account_data_dir: str, resource_groups: List[str]):
        urls_with_params = []
        if parameters:
            for parameter in parameters:
                param_names = parameter['Name']
                param_dynamic_value = parameter['Value']
                param_real_values = get_dynamic_values_from_files(param_dynamic_value, account_data_dir)

                for param_real_value in param_real_values:
                    modified_url = _url
                    zipped = zip(param_names.split(' '), param_real_value.split(' '))
                    for param, value in zipped:
                        modified_url = modified_url.replace('{{{0}}}'.format(param), value)

                    urls_with_params.append(modified_url)
        else:
            urls_with_params.append(_url)

        complete_urls = []

        for _url in urls_with_params:
            if '/{resourceGroupName}/' in _url:
                for resource_group in resource_groups:
                    complete_urls.append(_url.replace('{{{0}}}'.format('resourceGroupName'), resource_group))
            else:
                complete_urls.append(_url)

        return complete_urls

    def _get_results(self, base_url: str, headers: dict, parameters: List[dict], account_data_dir: str, resource_groups: List[str]) -> dict:
        results = {'value': []}
        urls = AzureScanner._build_urls(base_url, parameters, account_data_dir, resource_groups)
        for url in urls:
            logger.info(f'Invoking {url}')
            call_summary = {
                'request': url
            }
            response = invoke_get_request(url, headers, on_giveup=self._default_on_backoff_giveup)
            if response.status_code == 200:
                AzureScanner._concat_results(results, response)
            else:
                call_summary['error'] = json.loads(response.content.decode('utf-8'))['error']
                logger.error(self._parse_error(call_summary))
            self.summary.put_nowait(call_summary)
        results['urls'] = urls
        return results

    @staticmethod
    def _default_on_backoff_giveup(details: dict) -> None:
        logger.error('Given up on request for {args[0]} after {tries} tries'.format(**details))

    @staticmethod
    def _concat_results(results: dict, response: Response) -> None:
        if response.status_code == 200:
            result = json.loads(response.text)
            if 'value' in result:
                results['value'].extend(result['value'])
            else:
                if isinstance(result, list):
                    results['value'].extend(result)
                else:
                    results['value'].append(result)

    def _get_resource_groups(self, headers: dict) -> List[str]:
        url = f'https://management.azure.com/subscriptions/{self.subscription_id}/resourcegroups?api-version=2020-09-01'
        results = self._get_results(url, headers, [], self.account_data_dir, [])
        output_file = self._get_result_file_path(self.account_data_dir, 'resource-groups')
        self._save_result(results, output_file)
        logger.info(f'Results from {url} were saved to {output_file}')
        return get_dynamic_values_from_files('resource-groups.json|.value[].name', self.account_data_dir)

    @staticmethod
    def _add_resource_group(results: dict) -> None:
        for item in results['value']:
            if 'id' in item:
                item_id = item['id']
                try:
                    resource_group = item_id.split('resourceGroups/')[1].split('/')[0]
                    item['resourceGroup'] = resource_group
                except Exception:
                    pass

    @staticmethod
    def _get_result_file_path(account_data_dir: str, filename: str):
        return os.path.join(account_data_dir, filename + '.json')

    @staticmethod
    def _parse_error(call_summary: dict):
        error_code = call_summary['error']['code']
        error_msg = call_summary['error']['message']
        return f'{call_summary["request"]}: {error_code} - {error_msg}'



from typing import Optional

import boto3
from botocore.exceptions import ClientError

from dragoneye.utils.app_logger import logger
from dragoneye.dragoneye_exception import DragoneyeException


class AwsSessionFactory:
    @staticmethod
    def get_session(profile_name: Optional[str] = None, region: Optional[str] = None):
        start_msg = 'Will try to create a session'
        session_data = {}
        if region:
            session_data["region_name"] = region
        if profile_name:
            session_data["profile_name"] = profile_name
            start_msg = f'{start_msg} using profile {profile_name}'
        else:
            start_msg = f'{start_msg} using AWS auth-chain'
        logger.info(f'{start_msg}...')
        try:
            session = boto3.Session(**session_data)
        except ClientError as ex:
            raise DragoneyeException(f'An exception happened while trying to create session: {str(ex)}', ex.response['Error']['Message'])
        except Exception as ex:
            raise DragoneyeException(f'An unknown exception happened while trying to create session: {str(ex)}', str(ex))
        AwsSessionFactory.test_connectivity(session)
        logger.info('Session was created successfully')
        return session

    @staticmethod
    def get_session_using_assume_role(role_arn: str, external_id: str, region: Optional[str] = None, session_duration: int = 3600):
        role_session_name = "DragoneyeSession"
        logger.info('Will try to assume role using ARN: {} and external id {}...'.format(role_arn, external_id))
        try:
            client = boto3.client('sts')
            response = client.assume_role(RoleArn=role_arn,
                                          RoleSessionName=role_session_name,
                                          DurationSeconds=session_duration,
                                          ExternalId=external_id)
            credentials = response['Credentials']
            session_data = {
                "aws_access_key_id": credentials['AccessKeyId'],
                "aws_secret_access_key": credentials['SecretAccessKey'],
                "aws_session_token": credentials['SessionToken']
            }
            if region:
                session_data['region_name'] = region
            session = boto3.Session(**session_data)
        except ClientError as ex:
            raise DragoneyeException(f'An exception happened while trying to assume role: {str(ex)}', ex.response['Error']['Message'])
        except Exception as ex:
            raise DragoneyeException(f'An unknown exception happened while trying to assume role: {str(ex)}', str(ex))
        AwsSessionFactory.test_connectivity(session)
        logger.info('Session was created successfully')
        return session

    @staticmethod
    def test_connectivity(session):
        sts = session.client("sts")
        try:
            sts.get_caller_identity()
        except ClientError as ex:
            if "InvalidClientTokenId" in str(ex):
                raise DragoneyeException('sts.get_caller_identity failed with InvalidClientTokenId. '
                                         'Likely cause is no AWS credentials are set', str(ex))
            raise DragoneyeException('Unknown exception when trying to call sts.get_caller_identity: {}'.format(ex), str(ex))



import collections
import itertools
import json
import os
from functools import lru_cache
from typing import List, Deque, Optional
from googleapiclient.discovery import build
from googleapiclient.errors import HttpError

from dragoneye.cloud_scanner.base_cloud_scanner import BaseCloudScanner
from dragoneye.cloud_scanner.gcp.gcp_scan_settings import GcpCloudScanSettings
from dragoneye.config import config
from dragoneye.utils.misc_utils import elapsed_time, init_directory, custom_serializer, get_dynamic_values_from_files
from dragoneye.utils.threading_utils import ThreadedFunctionData, execute_parallel_functions_in_threads
from dragoneye.utils.app_logger import logger


class GcpScanner(BaseCloudScanner):
    def __init__(self, credentials, settings: GcpCloudScanSettings):
        super().__init__(settings)
        self.credentials = credentials
        self.services: list = []
        self.project_id = settings.project_id

    @elapsed_time('Scanning GCP live environment took {} seconds')
    def scan(self) -> str:
        self.account_data_dir = init_directory(self.settings.output_path, self.settings.account_name, self.settings.clean)

        dependent_commands, independent_commands = self._get_scan_commands()

        non_dependable_tasks: List[ThreadedFunctionData] = []
        dependable_tasks: List[ThreadedFunctionData] = []
        deque_tasks: Deque[List[ThreadedFunctionData]] = collections.deque()

        for independent_command in independent_commands:
            non_dependable_tasks.append(ThreadedFunctionData(
                self._execute_scan_commands,
                (independent_command,),
                'exception on command {}'.format(independent_command)))

        deque_tasks.append(non_dependable_tasks)

        for dependent_command in dependent_commands:
            dependable_tasks.append(ThreadedFunctionData(
                self._execute_scan_commands,
                (dependent_command,),
                'exception on command {}'.format(dependent_command)))

        for dependable_task in dependable_tasks:
            deque_tasks.append([dependable_task])
        execute_parallel_functions_in_threads(deque_tasks, config.get('MAX_WORKERS'))

        self._print_summary()

        for service in self.services:
            service.close()

        return os.path.abspath(os.path.join(self.account_data_dir, '..'))

    def _execute_scan_commands(self, scan_command: dict) -> None:
        service_name = scan_command['ServiceName']
        api_version = scan_command['ApiVersion']
        resource_type = scan_command['ResourceType']
        method = scan_command['Method']
        output_file = os.path.join(self.account_data_dir, f'{service_name}-{api_version}-{resource_type}-{method}.json')
        if os.path.isfile(output_file):
            # Data already scanned, so skip
            logger.warning('Response already present at {}'.format(output_file))
            return

        service = self._create_service(service_name, api_version)
        resource_response = getattr(service, resource_type)()

        all_parameters = self._get_parameters(scan_command, self.account_data_dir)
        all_items = []
        all_call_summary = []
        call_summary = {
            "service": service_name,
            "api_version": api_version,
            "resource_type": resource_type,
            'method': method
        }

        if all_parameters is not None:
            for parameters in all_parameters:
                updated_call_summary = call_summary.copy()
                updated_call_summary['parameters'] = parameters
                all_items.extend(self._get_results(updated_call_summary, resource_response))
                all_call_summary.append(updated_call_summary)
        else:
            updated_call_summary = call_summary.copy()
            updated_call_summary['parameters'] = {}
            all_items.extend(self._get_results(updated_call_summary, resource_response))
            all_call_summary.append(updated_call_summary)

        with open(output_file, "w") as file:
            json.dump({'value': all_items}, file, indent=4, default=custom_serializer)

        for call_summary in all_call_summary:
            self.summary.put_nowait(call_summary)
            if any(x in call_summary for x in ('error', 'exception')):
                logger.error(self._parse_error(call_summary))
            else:
                logger.info(f'Results from {self._get_call_representation(call_summary)} were saved to {output_file}')

    @lru_cache(maxsize=None)
    def _create_service(self, service_name: str, version: str):
        service = build(service_name, version, credentials=self.credentials)
        self.services.append(service)
        return service

    def _get_parameters(self, scan_command: dict, account_data_dir: str) -> Optional[List[dict]]:
        if not scan_command.get('Parameters'):
            return None  # No parameters required

        multi_params = []
        single_param_product = []
        single_param_data = {}
        multi_param_data = {}

        for parameter in scan_command.get('Parameters', []):
            param_names = parameter['Name']
            param_dynamic_value = parameter['Value']
            param_real_values = get_dynamic_values_from_files(param_dynamic_value, account_data_dir)
            # Multiple parameters from same object
            if ' ' in param_names:
                for param_real_value in param_real_values:
                    zipped = zip(param_names.split(' '), param_real_value.split(' '))
                    for param, value in zipped:
                        if param not in multi_param_data:
                            multi_param_data[param] = []
                        if value not in multi_param_data[param]:
                            multi_param_data[param].append(value)
            # One parameter from same object
            else:
                if '$project' in param_dynamic_value:
                    single_param_data[param_names] = [param_dynamic_value.replace('$project', self.project_id)]
                else:
                    single_param_data[param_names] = param_real_values

        keys = single_param_data.keys()
        for product in itertools.product(*single_param_data.values()):
            if product:
                single_param_product.append(dict(zip(keys, product)))

        multi_param_length = len(next(iter(multi_param_data.values()), []))
        for index in range(multi_param_length - 1):
            multi_param = {}
            for key, items in multi_param_data.items():
                multi_param[key] = items[index]
            multi_params.append(multi_param)

        if multi_params and single_param_product:
            stitched_params = []
            for multi_param in multi_params:
                for single_param in single_param_product:
                    multi_param_copy = multi_param.copy()
                    multi_param_copy.update(single_param)
                    stitched_params.append(multi_param_copy)
            return stitched_params
        else:
            return multi_params or single_param_product

    def _get_results(self, call_summary: dict, resource_response):
        all_items = []
        try:
            logger.info(f'Invoking {self._get_call_representation(call_summary)}')
            method_name = call_summary['method']
            method_name_next = f'{method_name}_next'
            method = getattr(resource_response, method_name)
            request = method(**call_summary['parameters'])

            while request is not None:
                response = request.execute()
                if 'list' in method_name or 'nextPageToken' in response:
                    for values in response.values():
                        if isinstance(values, list):
                            all_items.extend(values)
                            break

                    request = getattr(resource_response, method_name_next)(previous_request=request, previous_response=response)
                else:
                    if response:
                        all_items.append(response)
                    break
        except HttpError as ex:
            call_summary['error'] = json.loads(ex.content.decode('utf-8'))['error']
        except Exception as ex:
            call_summary['exception'] = str(ex)

        return all_items

    @staticmethod
    def _get_call_representation(call_summary: dict) -> str:
        parameters_text = ', '.join(f'{key}={value}' for key, value in call_summary['parameters'].items())
        return f'{call_summary["service"]}.{call_summary["api_version"]}.{call_summary["resource_type"]}.{call_summary["method"]}({parameters_text})'

    @staticmethod
    def _parse_error(call_summary: dict) -> str:
        if 'error' in call_summary:
            error_code = call_summary['error']['code']
            error_msg = call_summary['error']['message']
            return f'{GcpScanner._get_call_representation(call_summary)}: {error_code} - {error_msg}'
        else:
            return f'{GcpScanner._get_call_representation(call_summary)}: {call_summary["exception"]}'
